#include <stdio.h>
#include "tty.h"
#include "b.h"
#include "w.h"
#include "bw.h"
#include "menu.h"
#include "pw.h"
#include "qw.h"
#include "tw.h"
#include "ublock.h"
#include "mouse.h"

int xtm_isx;       /* whether the terminal is an xterm */
int xtm_active;    /* whether the -usemouse option is set */
int xtm_x1,xtm_y1; /* saved initial click position for xtmzone1() */
P   *xtm_pos1;     /* position struct for click position */

/* find the position struct and window associated to a click */

int findclickwin(bw,y,w)
BW *bw;
int y;
W **w;
{
  BW *bww;

  *w=bw->parent->t->topwin;
  do {
   if ((*w)->y!=-1)
    if (((*w)->y<y)&&((*w)->y+(*w)->h>=y)) {
      if ((*w)->watom->what&(TYPEPW+TYPETW)) {
        bww=(BW *)((*w)->object);
        if (bww->y>=y) return 0;
        if (bww->y+bww->h<y) return 0;
      }
      return (*w)->watom->what;
    }
   *w=(*w)->link.next;
  }
  while (*w!=(*w)->t->topwin);
  return 0;
}

P *findclickpos(bw,x,y)
BW *bw;
int x;
int y;
{
  P *p;
  int i;
  
  p=pdup(bw->top); 
  for (i=bw->y+1;i<y;i++) {
    if (p->line==bw->b->eof->line) return p;
    pnextl(p);
  }
  if (bw->parent->watom->what&TYPEPW) {
   x+=((PW *)bw->object)->promptofst;
   x-=((PW *)bw->object)->promptlen;
   if (x<1) x=1;
  }
  pcol(p,bw->offset+x-1);
  return p;
}

void xtmopen()   /* set mouseactive and initialize tracking mode */
{
/* must be called after term lookup (sets xtm_isx) and rc parsing complete
 * maybe later check that all callbacks are mapped properly in rc file...
 * for now we let (l)users nuke the term as they wish
 */
  if (!xtm_isx) { xtm_active=0; return; }
  xtm_pos1=(P *)0;
  xtm_active=1;
  ttputs("\033[?1001h");
  ttflsh();
}

void xtmclose()  /* end tracking mode, keep xtm_active unchanged */
{
  if (!xtm_isx) return;
  prm(xtm_pos1);
  ttputs("\033[?1001l");
  ttflsh();
}

int xtmnone(bw)   /* just parse a click and send reply */
BW *bw;
{
  ttgetc(); /* get x,y position we don't care about */
  ttgetc();
  ttputs("\033[0;0;0;0;0T"); /* reply we're not interested */
  ttputs("\033[?9s");        /* work around a stupid xterm bug */
  ttflsh();
  return 0; /* all was fine */
}

void domenuclick(m,x) /* emulate enter at x in menu */
MENU *m;
int x;
{
  x = m->top + x/(m->width+1);
  if (x<m->nitems) {
    m->cursor = x;
    umrtn(m);
  }
}

int xtmstart(bworg)  /* parse a click, send reply, move cursor, save pos */
BW *bworg;
{
  int rslt;
  char s[30];
  W *w;
  BW *bw;
  
  xtm_x1=ttgetc()-32;
  xtm_y1=ttgetc()-32;
  rslt=findclickwin(bworg,xtm_y1,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  if (rslt&(TYPETW+TYPEPW)) 
  {
    bw=(BW *)w->object;
    prm(xtm_pos1);
    xtm_pos1=findclickpos(bw,xtm_x1,xtm_y1);
    sprintf(s,"\033[1;%d;%d;%d;%dT\033[?9s",
      xtm_x1, xtm_y1, bw->y+1, bw->y+bw->h+1); /* track in the window range */
    ttputs(s);
    ttflsh();
    unmark(bw);
    pset(bw->cursor,xtm_pos1);           /* move cursor */
    bw->cursor->xcol=piscol(bw->cursor);
    return 0;
  } 
  else
  {
    ttputs("\033[0;0;0;0;0T"); /* no tracking needed */
    ttputs("\033[?9s");        /* work around a stupid xterm bug */
    ttflsh();
    /* click in menus */
    if (rslt&TYPEMENU) domenuclick((MENU *)w->object,xtm_x1);
    return 0;
  }  
}

int xtmclick(bworg)  /* parse a click, send reply, move cursor */
BW *bworg;
{
  int rslt,x,y;
  W *w;
  BW *bw;
  P *p;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  ttputs("\033[0;0;0;0;0T"); /* no tracking needed */
  ttputs("\033[?9s");        /* work around a stupid xterm bug */
  ttflsh();
  rslt=findclickwin(bworg,y,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  if (rslt&(TYPETW+TYPEPW)) 
  {
    bw=(BW *)w->object;
    p=findclickpos(bw,x,y);
    pset(bw->cursor,p);           /* move cursor */
    prm(p);
    bw->cursor->xcol=piscol(bw->cursor);
  }
  if (rslt&TYPEMENU) domenuclick((MENU *)w->object,x);
  return 0;
}

int xtmcopy(bworg)   /* parse a click, send reply, copy */                 
BW *bworg;
{
  int rslt,x,y;
  W *w;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,y,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  ttputs("\033[0;0;0;0;0T");   /* no tracking needed */
  ttputs("\033[?9s");          /* work around a stupid xterm bug */
  ttflsh();
  if (rslt&TYPETW)
    if (((BW *)(w->object))->b->rdonly) {
      msgnw((BW *)w->object,"Read only");
      if (beep) ttputc(7);
      return 0;
    }
  if (rslt&(TYPETW+TYPEPW))
    ublkcpy((BW *)w->object);
  return 0;
}

int xtmmove(bworg)   /* parse a click, send reply, move */                 
BW *bworg;
{
  int rslt,x,y;
  W *w;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,y,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  ttputs("\033[0;0;0;0;0T");   /* no tracking needed */
  ttputs("\033[?9s");          /* work around a stupid xterm bug */
  ttflsh();
  if (rslt&TYPETW)
    if (((BW *)(w->object))->b->rdonly) {
      msgnw((BW *)w->object,"Read only");
      if (beep) ttputc(7);
      return 0;
    }
  if (rslt&(TYPETW+TYPEPW)) 
    ublkmove((BW *)w->object);
  return 0;
}

int xtmcopyat(bworg) /* parse a click, send reply, move cursor, copy */    
BW *bworg;
{
  int rslt,x,y;
  W *w;
  BW *bw;
  P *p;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,y,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  ttputs("\033[0;0;0;0;0T");   /* no tracking needed */
  ttputs("\033[?9s");          /* work around a stupid xterm bug */
  ttflsh();
  if (rslt&(TYPETW+TYPEPW)) 
  {
    bw=(BW *)w->object;
    p=findclickpos(bw,x,y);
    pset(bw->cursor,p);           /* move cursor */
    prm(p);
    bw->cursor->xcol=piscol(bw->cursor);
    if ((rslt&TYPETW)&&(bw->b->rdonly)) {
      msgnw(bw,"Read only");
      if (beep) ttputc(7);
      return 0;
    }
    ublkcpy(bw);
  }
  return 0;
}

int xtmmoveat(bworg) /* parse a click, send reply, move cursor, move */   
BW *bworg;
{
  int rslt,x,y;
  W *w;
  BW *bw;
  P *p;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,y,&w);
  if (rslt!=0) w->t->curwin=w; /* make window active */
  ttputs("\033[0;0;0;0;0T");   /* no tracking needed */
  ttputs("\033[?9s");          /* work around a stupid xterm bug */
  ttflsh();
  if (rslt&(TYPETW+TYPEPW)) 
  {
    bw=(BW *)w->object;
    p=findclickpos(bw,x,y);
    pset(bw->cursor,p);           /* move cursor */
    prm(p);
    bw->cursor->xcol=piscol(bw->cursor);
    if ((rslt&TYPETW)&&(bw->b->rdonly)) {
      msgnw(bw,"Read only");
      if (beep) ttputc(7);
      return 0;
    }
    ublkmove(bw);
  } 
  return 0;
}

int xtmzone1(bworg)  /* parse a track-end, move cursor, set block */      
BW *bworg;
{
  int rslt,x,y;
  W *w1,*w2;
  BW *bw;
  P *p;
  
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,xtm_y1,&w1);
  if (!(rslt&(TYPETW+TYPEPW))) return 0; /* cases we can't deal with */
  if (rslt!=findclickwin(bworg,(x==1)?(y-1):y,&w2)) return 0;
  if (w1!=w2) return 0;
  w1->t->curwin=w1;                     /* make window active */

  bw=(BW *)w1->object;
  prm(xtm_pos1);
  xtm_pos1=findclickpos(bw,xtm_x1,xtm_y1);
  pset(bw->cursor,xtm_pos1);           /* move cursor */
  bw->cursor->xcol=piscol(bw->cursor);
  umarkb(bw);
  p=findclickpos(bw,x,y);
  pset(bw->cursor,p);
  bw->cursor->xcol=piscol(bw->cursor);
  prm(p);
  umarkk(bw);
  return 0;
}

int xtmzone2(bworg)  /* parse a track-end, move cursor, set block */      
BW *bworg;
{
  int rslt,x2,y2,x,y;
  W *w1,*w2;
  BW *bw;
  P *p;
  
  xtm_x1=ttgetc()-32;
  xtm_y1=ttgetc()-32;
  x2=ttgetc()-32;
  y2=ttgetc()-32;
  x=ttgetc()-32;
  y=ttgetc()-32;
  rslt=findclickwin(bworg,xtm_y1,&w1);
  if (!(rslt&(TYPETW+TYPEPW))) return 0; /* cases we can't deal with */
  if (rslt!=findclickwin(bworg,(x2==1)?(y2-1):y2,&w2)) return 0;
  if (w1!=w2) return 0;
  w1->t->curwin=w1;                     /* make window active */

  bw=(BW *)w1->object;
  prm(xtm_pos1);
  xtm_pos1=findclickpos(bw,xtm_x1,xtm_y1);
  pset(bw->cursor,xtm_pos1);           /* move cursor */
  bw->cursor->xcol=piscol(bw->cursor);
  umarkb(bw);
  p=findclickpos(bw,x2,y2);
  pset(bw->cursor,p);
  bw->cursor->xcol=piscol(bw->cursor);
  prm(p);
  umarkk(bw);
  if ((y-xtm_y1<y2-y)||(xtm_y1==y2)&&(x-xtm_x1<x2-x)) 
  {  /* definitely a backwards selection */
    pset(bw->cursor,xtm_pos1);
    bw->cursor->xcol=piscol(bw->cursor);
  }
  return 0;
}

